﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using SharpDX;
using SharpDX.Toolkit;
using SharpDX.Toolkit.Graphics;

namespace DeferredRenderer
{
    class DeferredRenderer
    {
        // A reference to the device from game
        GraphicsDevice m_device;

        // The deep frame buffer textures
        RenderTarget2D m_RTColor;
        RenderTarget2D m_RTNormal;
        RenderTarget2D m_RTPosition;
        SharpDX.Direct3D11.RenderTargetView[] m_deepFrameBuffer;

        // Standard texture sampling for the geometry pass
        SamplerState m_linearSamplerState;
        // Sampling for the "post process effects"
        SamplerState m_pointSamplerState;

        private Effect m_geometryEffect;    // Effect to render standard geometry
        private Effect m_dirLightEffect;    // Effect to copy the ambient lightening and compute large lights
        private Effect m_pointLightEffect;  // Effect to render one point light

        BlendState m_blendStateOpaque;      // BlendState for normal geometry
        BlendState m_blendStateAdditive;    // BLendState for the rendering of lights
        RasterizerState m_backfaceCullingState;
        RasterizerState m_scissorState;

        public Effect GetStdGeomtryEffect() { return m_geometryEffect; }

        // Just some lights for testing
        PointLight[] m_pointLights;

        /// <summary>
        /// Create the deep frame buffer and load the shaders...
        /// </summary>
        public DeferredRenderer(GraphicsDevice _device)
        {
            m_device = _device;
            CreateFrameBuffer();

            // Create linear and point sampler
            var samplerStateDesc = SharpDX.Direct3D11.SamplerStateDescription.Default();
            samplerStateDesc.AddressV = SharpDX.Direct3D11.TextureAddressMode.Wrap;
            samplerStateDesc.AddressU = SharpDX.Direct3D11.TextureAddressMode.Wrap;
            m_linearSamplerState = SamplerState.New(m_device, "LinearSampler", samplerStateDesc);

            samplerStateDesc.Filter = SharpDX.Direct3D11.Filter.MinMagMipPoint;
            m_pointSamplerState = SamplerState.New(m_device, "PointSampler", samplerStateDesc);

            // Load all the shaders for the different passes
            m_geometryEffect = Helpers.LoadEffect(_device, "geometry.fx");
            m_geometryEffect.Parameters["diffuseSampler"].SetResource(m_linearSamplerState);

            m_dirLightEffect = Helpers.LoadEffect(_device, "basic_lights.fx");
            m_dirLightEffect.Parameters["pointSampler"].SetResource(m_pointSamplerState);

            m_pointLightEffect = Helpers.LoadEffect(_device, "pointlight.fx");
            m_pointLightEffect.Parameters["pointSampler"].SetResource(m_pointSamplerState);

            // States for geometry and light rendering
            var blendStateDesc = SharpDX.Direct3D11.BlendStateDescription.Default();
            m_blendStateOpaque = BlendState.New(_device, "Opaque", blendStateDesc);
            blendStateDesc.RenderTarget[0].DestinationBlend = SharpDX.Direct3D11.BlendOption.One;
            blendStateDesc.RenderTarget[0].SourceBlend = SharpDX.Direct3D11.BlendOption.One;
            blendStateDesc.RenderTarget[0].BlendOperation = SharpDX.Direct3D11.BlendOperation.Add;
            blendStateDesc.RenderTarget[0].IsBlendEnabled = true;
            m_blendStateAdditive = BlendState.New(_device, "Additive", blendStateDesc);

            var rasterizerStateDesc = SharpDX.Direct3D11.RasterizerStateDescription.Default();
            m_backfaceCullingState = RasterizerState.New(_device, "CullModeBack", rasterizerStateDesc);
            rasterizerStateDesc.IsScissorEnabled = true;
            m_scissorState = RasterizerState.New(_device, "Scissor", rasterizerStateDesc);
        }

        /// <summary>
        /// Create Color, Normals and Position targets
        /// </summary>
        public void CreateFrameBuffer()
        {
            m_RTColor = RenderTarget2D.New(m_device, m_device.BackBuffer.Width, m_device.BackBuffer.Height, SharpDX.DXGI.Format.R8G8B8A8_UNorm);
            m_RTNormal = RenderTarget2D.New(m_device, m_device.BackBuffer.Width, m_device.BackBuffer.Height, SharpDX.DXGI.Format.R11G11B10_Float);
            m_RTPosition = RenderTarget2D.New(m_device, m_device.BackBuffer.Width, m_device.BackBuffer.Height, SharpDX.DXGI.Format.R16G16B16A16_Float);

            m_deepFrameBuffer = new SharpDX.Direct3D11.RenderTargetView[3] {
                    m_RTColor.RenderTargetView[ViewType.Single, 0, 0],
                    m_RTNormal.RenderTargetView[ViewType.Single, 0, 0],
                    m_RTPosition.RenderTargetView[ViewType.Single, 0, 0]};
        }

        // Creates the desired number of test lights
        public void CreateSomeTestLights(int _num)
        {
            Random rnd = new Random();
            m_pointLights = new PointLight[_num];
            for (int i = 0; i < _num; ++i)
            {
                m_pointLights[i].Set( 1.0f, 5.0f + rnd.NextFloat(-1.0f, 1.0f));
                m_pointLights[i].pos = new Vector3(rnd.NextFloat(-15.0f, 15.0f), rnd.NextFloat(0.0f, 80.0f), rnd.NextFloat(-15.0f, 15.0f));
                m_pointLights[i].color = new Vector3(1.0f, 0.8f, 0.5f);
            }
        }

        /// <summary>
        /// Rotate the test lights a little bit
        /// </summary>
        /// <param name="_dt">delta time in seconds</param>
        public void UpdateTestLights(float _dt)
        {
            Matrix rotate = Matrix.RotationY(_dt * 0.2f);

            for (int i = 0; i < m_pointLights.Length; ++i )
            {
                m_pointLights[i].pos = Vector3.TransformCoordinate(m_pointLights[i].pos, rotate);
            }
        }

        /// <summary>
        /// Set all render targets and the shader to render to the deep frame buffer.
        /// </summary>
        public void BeginGeometryPass()
        {
            // Important: all render targets and the depth buffer must have the
            // same size. Since our RTs have the back buffer size the standard
            // depth buffer is compatible.
            // Hint: A good optimization would be to recompute position from depth
            // and not to render to m_RTPosition!
            m_device.SetRenderTargets(m_device.DepthStencilBuffer, m_deepFrameBuffer);

            // Clear all targets
            m_device.Clear(m_deepFrameBuffer[0], new Color(5, 14, 30));
            m_device.Clear(m_deepFrameBuffer[1], new Color(0, 0, 0));
            m_device.Clear(m_deepFrameBuffer[2], new Color(0, 0, 0));

            m_device.SetBlendState(m_blendStateOpaque);
            m_device.SetRasterizerState(m_backfaceCullingState);
            m_geometryEffect.CurrentTechnique.Passes[0].Apply();
        }

        /// <summary>
        /// Set the old back buffer and do ambient+directional lightening
        /// </summary>
        public void EndGeomtryPass()
        {
            var lightCB = m_dirLightEffect.ConstantBuffers["Light"];
            lightCB.Parameters["lightDir"].SetValue(new Vector3(0.707f, 0.707f, 0.0f));
            lightCB.Parameters["ambientLight"].SetValue(0.1f);
            lightCB.Parameters["dirLightColor"].SetValue(new Vector3(0.2f, 0.2f, 0.4f));

            // Draw to real back-buffer
            m_device.SetRenderTargets(m_device.BackBuffer);

            // Set copy effect
            m_dirLightEffect.CurrentTechnique.Passes[0].Apply();
            m_dirLightEffect.Parameters["colorTexture"].SetResource(m_RTColor.ShaderResourceView[ViewType.Single, 0, 0]);
            m_dirLightEffect.Parameters["normalTexture"].SetResource(m_RTNormal.ShaderResourceView[ViewType.Single, 0, 0]);
            m_device.DrawQuad();
        }

        /// <summary>
        /// Draw all the many lights
        /// </summary>
        public void LightPass(Matrix _viewProjection)
        {
            m_device.SetBlendState(m_blendStateAdditive);
            m_device.SetRasterizerState(m_scissorState);

            // Set effect and read from deep frame buffer
            m_pointLightEffect.Parameters["colorTexture"].SetResource(m_RTColor.ShaderResourceView[ViewType.Single, 0, 0]);
            m_pointLightEffect.Parameters["normalTexture"].SetResource(m_RTNormal.ShaderResourceView[ViewType.Single, 0, 0]);
            m_pointLightEffect.Parameters["positionTexture"].SetResource(m_RTPosition.ShaderResourceView[ViewType.Single, 0, 0]);

            var lightCB = m_pointLightEffect.ConstantBuffers["Lights"];

            // Resolution is required to compute the correct scissor rect for optimizations
            int screenWidth = m_RTColor.Width;
            int screenHeight = m_RTColor.Height;
            float aspect = screenWidth / (float)screenHeight;

            foreach(PointLight light in m_pointLights)
            {
                lightCB.Parameters["light"].SetValue(light);
                m_pointLightEffect.CurrentTechnique.Passes[0].Apply();
                // Compute where in the screen the light is.
                Vector4 projectedPos = Vector3.Transform(light.pos, _viewProjection);
                float x = 0.5f + 0.5f * projectedPos.X / projectedPos.W;
                float y = 0.5f - 0.5f * projectedPos.Y / projectedPos.W;
                float r = light.Range() / projectedPos.W;
                m_device.SetScissorRectangles((int)((x - r) * screenWidth), (int)((y - r * aspect) * screenHeight), (int)((x + r) * screenWidth), (int)((y + r * aspect) * screenHeight));
                m_device.DrawQuad();
            }
        }




        struct PointLight
        {
            public Vector3 pos;
            float att1;				// -1 / (outer_range-inner_range)
            public Vector3 color;
            float att2;				// 1 + inner_range / (outer_range-inner_range)
            // The inner_range is the distance at which the light starts to fade out.
            // The outer_range is the distance where the light gets 0.

            public void Set(float innerRange, float outerRange)
            {
                att1 = -1 / (outerRange - innerRange);
                att2 = 1 - innerRange * att1;
            }

            // Return the outerRange
            public float Range()
            {
                // Must be recomputed from att1 and att2
                float innerRange = (1 - att2) / att1;
                return -1 / att1 + innerRange;
            }
        }

    }
}
