﻿using SharpDX;
using SharpDX.Toolkit;
using SharpDX.Toolkit.Graphics;
using Assimp;
using System;

namespace DeferredRenderer
{
    class DrawModel : Game
    {
        private GraphicsDeviceManager m_graphicsDeviceManager;
        private DeferredRenderer m_render;
        private Model m_model;
        private Matrix m_viewProjection;
        private Vector3 m_cameraPosition = new Vector3(0.0f, 40.0f, -90.0f);

        DepthStencilState m_depthStencilStateState;

        /// <summary>
        /// Entry point create and starts the game
        /// </summary>
        static void Main()
        {
            var program = new DrawModel();
            program.Run();
        }

        DrawModel()
        {
            // Creates a graphics manager. This is mandatory.
            m_graphicsDeviceManager = new GraphicsDeviceManager(this);
            m_graphicsDeviceManager.PreferredDepthStencilFormat = DepthFormat.Depth32; 
            m_graphicsDeviceManager.SetPreferredGraphicsProfile(SharpDX.Direct3D.FeatureLevel.Level_10_1);

            // Setup the relative directory to the executable directory
            // for loading contents with the ContentManager
            Content.RootDirectory = "content";
        }

        protected override void Initialize()
        {
            base.Initialize();

            Window.Title = "Deferred Rendering Demo";

            // Create renderer in initialize - otherwise the device is not created
            m_render = new DeferredRenderer(m_graphicsDeviceManager.GraphicsDevice);
            m_render.CreateSomeTestLights(2000);

            // Setup Camera
            Matrix view = Matrix.LookAtRH(
                m_cameraPosition,                  // Position
                new Vector3(0.0f, 20.0f, 0.0f),     // At (point which is centered in the middle of the screen).
                new Vector3(0.0f, 1.0f, 0.0f));     // Up
            Matrix projection = Matrix.PerspectiveFovRH(
                0.6f,                               // Field of view
                (float)GraphicsDevice.BackBuffer.Width / GraphicsDevice.BackBuffer.Height,  // Aspect ratio
                0.5f,                               // Near clipping plane
                200.0f);                            // Far clipping plane
            m_viewProjection = view * projection;

            var depthStencilStateDesc = SharpDX.Direct3D11.DepthStencilStateDescription.Default();
            m_depthStencilStateState = DepthStencilState.New(GraphicsDevice, "NormalZBufferUse", depthStencilStateDesc);
        }

        protected override void LoadContent()
        {
            // Importer for many models
            var importer = new AssimpImporter();

            // Load a specific model
            string fileName = System.IO.Path.GetFullPath(Content.RootDirectory + "/tower.3ds");
            Scene scene = importer.ImportFile(fileName, PostProcessSteps.MakeLeftHanded);
            m_model = new Model(scene, GraphicsDevice, Content);
        }

        protected override void Draw(GameTime gameTime)
        {
            // Clears the screen
            GraphicsDevice.Clear(new Color(255, 144, 1));

            GraphicsDevice.SetDepthStencilState(m_depthStencilStateState);

            // Do more or less the standard rendering but with shaders to fill
            // the MRT
            m_render.BeginGeometryPass();

            // Fill the one constant buffer. Hint: it is not necessary to set
            // things which did not change each frame. But in our case everything
            // is changing
            UpdateMatrices(gameTime, m_render.GetStdGeomtryEffect());

            // Draw model
            m_model.Draw(GraphicsDevice, m_render.GetStdGeomtryEffect());

            // The effectively makes things visible again: MRT -> Backbuffer
            m_render.EndGeomtryPass();

            // Draw all small dynamic lights
            m_render.UpdateTestLights(gameTime.ElapsedGameTime.Milliseconds * 0.001f);
            m_render.LightPass(m_viewProjection);

            base.Draw(gameTime);
        }

        private void UpdateMatrices(GameTime gameTime, Effect _effect)
        {
            // New world matrix
            Matrix world = Matrix.RotationY(0.8f);
            //world *= Matrix.Scaling(12, 12, 12);

            var transformCB = _effect.ConstantBuffers["Transforms"];
            transformCB.Parameters["worldViewProj"].SetValue(world * m_viewProjection);
            transformCB.Parameters["world"].SetValue(world);
            Matrix worldInvTr = Helpers.CreateInverseTranspose(ref world);
            transformCB.Parameters["worldInvTranspose"].SetValue(worldInvTr);
            transformCB.Parameters["cameraPos"].SetValue(m_cameraPosition);
        }

     /*   private void UpdateLights(GameTime gameTime, Matrix world)
        {
            var lightCB = m_simpleEffect.ConstantBuffers["Lights"];
            lightCB.Parameters["lightDir"].SetValue(new Vector3(0.707f, 0.707f, 0.0f));
            lightCB.Parameters["specularPower"].SetValue(4.0f);
            lightCB.Parameters["dirLightColor"].SetValue(new Vector3(0.1f, 0.1f, 0.2f));
            lightCB.Parameters["numPointLights"].SetValue(2);

            PointLight[] pointLights = new PointLight[2];
            pointLights[0].Set(3.0f, 10.0f + 0.5f * (float)Math.Sin(gameTime.TotalGameTime.TotalMilliseconds / 100));
            pointLights[1].Set(3.0f, 10.0f + 0.5f * (float)Math.Cos(gameTime.TotalGameTime.TotalMilliseconds / 100));
            pointLights[0].pos = Vector3.TransformCoordinate(new Vector3(10.0f, 1.0f, -1.5f), world);
            pointLights[1].pos = Vector3.TransformCoordinate(new Vector3(7.5f, 1.0f, -7.5f), world);
            pointLights[0].color = new Vector3( 1.0f, 0.8f, 0.5f );
            pointLights[1].color = new Vector3( 1.0f, 0.8f, 0.5f );

            lightCB.Parameters["lights"].SetValue(pointLights);
        }*/
    }
}
